@@ -27,17 +27,19 @@
 from __future__ import absolute_import
 
 import os
+import mock
 
 if os.environ.get('USE_TWISTED', False):
 
     from twisted.internet.defer import inlineCallbacks, Deferred, returnValue
     from twisted.internet.defer import succeed, DeferredList
     from twisted.trial import unittest
+    import twisted
     from six import PY3
 
     from autobahn import util
     from autobahn.twisted.wamp import ApplicationSession
-    from autobahn.wamp import message, role, serializer, types
+    from autobahn.wamp import message, role, serializer, types, uri
     from autobahn.wamp.exception import ApplicationError, NotAuthorized
     from autobahn.wamp.exception import InvalidUri, ProtocolError
 
@@ -65,6 +67,11 @@ if os.environ.get('USE_TWISTED', False):
             self._handler.onMessage(msg)
             self._fake_router_session = ApplicationSession()
 
+        def drop_registration(self, reg_id):
+            self._handler.onMessage(
+                message.Unregistered(0, reg_id)
+            )
+
         def send(self, msg):
             if self._log:
                 payload, isbinary = self._serializer.serialize(msg)
@@ -167,6 +174,47 @@ if os.environ.get('USE_TWISTED', False):
             # test-case was written
             handler.onClose(False)
 
+    class TestRegisterDecorator(unittest.TestCase):
+
+        def test_prefix(self):
+
+            class Prefix(ApplicationSession):
+
+                @uri.register(u'method_name')
+                def some_method(self):
+                    pass
+            session = Prefix()
+
+            session._transport = mock.Mock()
+            session.register(session, prefix=u"com.example.prefix.")
+
+            # we should have registered one method, with the prefix
+            # put in front
+            self.assertEqual(1, len(session._transport.mock_calls))
+            call = session._transport.mock_calls[0]
+            self.assertEqual("send", call[0])
+            reg = call.call_list()[0][1][0]
+            self.assertEqual(u"com.example.prefix.method_name", reg.procedure)
+
+        def test_auto_name(self):
+
+            class Magic(ApplicationSession):
+
+                @uri.register(None)
+                def some_method(self):
+                    pass
+            session = Magic()
+
+            session._transport = mock.Mock()
+            session.register(session, prefix=u"com.example.")
+
+            # Should auto-discover name by passing uri=None above
+            self.assertEqual(1, len(session._transport.mock_calls))
+            call = session._transport.mock_calls[0]
+            self.assertEqual("send", call[0])
+            reg = call.call_list()[0][1][0]
+            self.assertEqual(u"com.example.some_method", reg.procedure)
+
     class TestPublisher(unittest.TestCase):
 
         @inlineCallbacks
@@ -594,6 +642,34 @@ if os.environ.get('USE_TWISTED', False):
             registration = yield handler.register(on_call, u'com.myapp.procedure1')
             yield registration.unregister()
 
+        def test_unregister_no_such_registration(self):
+            if twisted.version.major < 14:
+                return
+            handler = ApplicationSession()
+            transport = MockTransport(handler)
+
+            with self.assertRaises(ProtocolError) as ctx:
+                transport.send(
+                    message.Unregister(0, 1234)
+                )
+            self.assertIn(
+                "UNREGISTERED received for non-existant registration",
+                str(ctx.exception)
+            )
+
+        @inlineCallbacks
+        def test_unregister_log(self):
+            handler = ApplicationSession()
+            transport = MockTransport(handler)
+
+            def on_call(*args, **kwargs):
+                on_call.called = True
+            on_call.called = False
+
+            registration = yield handler.register(on_call, u'com.myapp.procedure1')
+            transport.drop_registration(registration.id)
+            self.assertFalse(on_call.called)
+
         def test_on_disconnect_error(self):
             errors = []
 
